/*
 * Copyright(c) 2012-2018 Intel Corporation
 * SPDX-License-Identifier: BSD-3-Clause-Clear
 */

#ifndef ENGINE_BF_H_
#define ENGINE_BF_H_

void ocf_engine_backfill(struct ocf_request *req);

#endif /* ENGINE_BF_H_ */
