/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(C) 2019 Marvell International Ltd.
 */
#ifndef _OTX2_EP_VF_H_
#define _OTX2_EP_VF_H_

int
sdp_vf_setup_device(struct sdp_device *sdpvf);

#endif /*_OTX2_EP_VF_H_ */
