/*
 * Copyright(c) 2012-2018 Intel Corporation
 * SPDX-License-Identifier: BSD-3-Clause-Clear
 */

#ifndef ENGINE_WI_H_
#define ENGINE_WI_H_

int ocf_write_wi(struct ocf_request *req);

#endif /* ENGINE_WI_H_ */
