/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright (c) 2015-2020 Amazon.com, Inc. or its affiliates.
 * All rights reserved.
 */

#define	ENA_GEN_DATE	"Wed Sep 25 11:32:57 UTC 2019"
#define	ENA_GEN_COMMIT	"952697a9e0d3"
