/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2014-2020 Broadcom
 * All rights reserved.
 */

#ifndef ULP_HDR_FIELD_ENUMS_H_
#define ULP_HDR_FIELD_ENUMS_H_

enum bnxt_ulp_hf0 {
	BNXT_ULP_HF0_IDX_SVIF_INDEX              = 0,
	BNXT_ULP_HF0_IDX_O_ETH_DMAC              = 1,
	BNXT_ULP_HF0_IDX_O_ETH_SMAC              = 2,
	BNXT_ULP_HF0_IDX_O_ETH_TYPE              = 3,
	BNXT_ULP_HF0_IDX_OO_VLAN_CFI_PRI         = 4,
	BNXT_ULP_HF0_IDX_OO_VLAN_VID             = 5,
	BNXT_ULP_HF0_IDX_OO_VLAN_TYPE            = 6,
	BNXT_ULP_HF0_IDX_OI_VLAN_CFI_PRI         = 7,
	BNXT_ULP_HF0_IDX_OI_VLAN_VID             = 8,
	BNXT_ULP_HF0_IDX_OI_VLAN_TYPE            = 9,
	BNXT_ULP_HF0_IDX_O_IPV4_VER              = 10,
	BNXT_ULP_HF0_IDX_O_IPV4_TOS              = 11,
	BNXT_ULP_HF0_IDX_O_IPV4_LEN              = 12,
	BNXT_ULP_HF0_IDX_O_IPV4_FRAG_ID          = 13,
	BNXT_ULP_HF0_IDX_O_IPV4_FRAG_OFF         = 14,
	BNXT_ULP_HF0_IDX_O_IPV4_TTL              = 15,
	BNXT_ULP_HF0_IDX_O_IPV4_NEXT_PID         = 16,
	BNXT_ULP_HF0_IDX_O_IPV4_CSUM             = 17,
	BNXT_ULP_HF0_IDX_O_IPV4_SRC_ADDR         = 18,
	BNXT_ULP_HF0_IDX_O_IPV4_DST_ADDR         = 19,
	BNXT_ULP_HF0_IDX_O_UDP_SRC_PORT          = 20,
	BNXT_ULP_HF0_IDX_O_UDP_DST_PORT          = 21,
	BNXT_ULP_HF0_IDX_O_UDP_LENGTH            = 22,
	BNXT_ULP_HF0_IDX_O_UDP_CSUM              = 23
};

enum bnxt_ulp_hf_bitmask0 {
	BNXT_ULP_HF0_BITMASK_SVIF_INDEX          = 0x8000000000000000,
	BNXT_ULP_HF0_BITMASK_O_ETH_DMAC          = 0x4000000000000000,
	BNXT_ULP_HF0_BITMASK_O_ETH_SMAC          = 0x2000000000000000,
	BNXT_ULP_HF0_BITMASK_O_ETH_TYPE          = 0x1000000000000000,
	BNXT_ULP_HF0_BITMASK_OO_VLAN_CFI_PRI     = 0x0800000000000000,
	BNXT_ULP_HF0_BITMASK_OO_VLAN_VID         = 0x0400000000000000,
	BNXT_ULP_HF0_BITMASK_OO_VLAN_TYPE        = 0x0200000000000000,
	BNXT_ULP_HF0_BITMASK_OI_VLAN_CFI_PRI     = 0x0100000000000000,
	BNXT_ULP_HF0_BITMASK_OI_VLAN_VID         = 0x0080000000000000,
	BNXT_ULP_HF0_BITMASK_OI_VLAN_TYPE        = 0x0040000000000000,
	BNXT_ULP_HF0_BITMASK_O_IPV4_VER          = 0x0020000000000000,
	BNXT_ULP_HF0_BITMASK_O_IPV4_TOS          = 0x0010000000000000,
	BNXT_ULP_HF0_BITMASK_O_IPV4_LEN          = 0x0008000000000000,
	BNXT_ULP_HF0_BITMASK_O_IPV4_FRAG_ID      = 0x0004000000000000,
	BNXT_ULP_HF0_BITMASK_O_IPV4_FRAG_OFF     = 0x0002000000000000,
	BNXT_ULP_HF0_BITMASK_O_IPV4_TTL          = 0x0001000000000000,
	BNXT_ULP_HF0_BITMASK_O_IPV4_NEXT_PID     = 0x0000800000000000,
	BNXT_ULP_HF0_BITMASK_O_IPV4_CSUM         = 0x0000400000000000,
	BNXT_ULP_HF0_BITMASK_O_IPV4_SRC_ADDR     = 0x0000200000000000,
	BNXT_ULP_HF0_BITMASK_O_IPV4_DST_ADDR     = 0x0000100000000000,
	BNXT_ULP_HF0_BITMASK_O_UDP_SRC_PORT      = 0x0000080000000000,
	BNXT_ULP_HF0_BITMASK_O_UDP_DST_PORT      = 0x0000040000000000,
	BNXT_ULP_HF0_BITMASK_O_UDP_LENGTH        = 0x0000020000000000,
	BNXT_ULP_HF0_BITMASK_O_UDP_CSUM          = 0x0000010000000000
};

#endif /* _ULP_HDR_FIELD_ENUMS_H_ */
