/*
 * Copyright(c) 2012-2018 Intel Corporation
 * SPDX-License-Identifier: BSD-3-Clause-Clear
 */

#ifndef ENGINE_WA_H_
#define ENGINE_WA_H_

int ocf_write_wa(struct ocf_request *req);

#endif /* ENGINE_WA_H_ */
