/*
 * Copyright(c) 2012-2018 Intel Corporation
 * SPDX-License-Identifier: BSD-3-Clause-Clear
 */

#ifndef __OCF_MNGT_CORE_PRIV_H__
#define __OCF_MNGT_CORE_PRIV_H__

#include "../ocf_core_priv.h"

int ocf_mngt_core_init_front_volume(ocf_core_t core);

void ocf_mngt_core_clear_uuid_metadata(ocf_core_t core);

#endif /* __OCF_MNGT_CORE_PRIV_H__ */
