/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright (c) 2016 - 2018 Cavium Inc.
 * All rights reserved.
 * www.cavium.com
 */

#ifndef GTT_REG_ADDR_H
#define GTT_REG_ADDR_H

/* Win 2 */
//Access:RW   DataWidth:0x20   //
#define GTT_BAR0_MAP_REG_IGU_CMD                                      0x00f000UL

/* Win 3 */
//Access:RW   DataWidth:0x20   //
#define GTT_BAR0_MAP_REG_TSDM_RAM                                     0x010000UL

/* Win 4 */
//Access:RW   DataWidth:0x20   //
#define GTT_BAR0_MAP_REG_MSDM_RAM                                     0x011000UL

/* Win 5 */
//Access:RW   DataWidth:0x20   //
#define GTT_BAR0_MAP_REG_MSDM_RAM_1024                                0x012000UL

/* Win 6 */
//Access:RW   DataWidth:0x20   //
#define GTT_BAR0_MAP_REG_MSDM_RAM_2048                                0x013000UL

/* Win 7 */
//Access:RW   DataWidth:0x20   //
#define GTT_BAR0_MAP_REG_USDM_RAM                                     0x014000UL

/* Win 8 */
//Access:RW   DataWidth:0x20   //
#define GTT_BAR0_MAP_REG_USDM_RAM_1024                                0x015000UL

/* Win 9 */
//Access:RW   DataWidth:0x20   //
#define GTT_BAR0_MAP_REG_USDM_RAM_2048                                0x016000UL

/* Win 10 */
//Access:RW   DataWidth:0x20   //
#define GTT_BAR0_MAP_REG_XSDM_RAM                                     0x017000UL

/* Win 11 */
//Access:RW   DataWidth:0x20   //
#define GTT_BAR0_MAP_REG_XSDM_RAM_1024                                0x018000UL

/* Win 12 */
//Access:RW   DataWidth:0x20   //
#define GTT_BAR0_MAP_REG_YSDM_RAM                                     0x019000UL

/* Win 13 */
//Access:RW   DataWidth:0x20   //
#define GTT_BAR0_MAP_REG_PSDM_RAM                                     0x01a000UL

/* Win 14 */

#endif
